# distutils: language = c++
# distutils: include_dirs = src
# distutils: sources = src/voro++.cc

from __future__ import division

from libcpp.vector cimport vector
from libcpp cimport bool as cbool
from cython.operator cimport dereference

cdef extern from "voro++.hh" namespace "voro":
    cdef cppclass container_base:
        pass
    
    cdef cppclass container:
        container(double,double,double,double,double,double,
                int,int,int,cbool,cbool,cbool,int) except +
        cbool compute_cell(voronoicell_neighbor &c,c_loop_all &vl)
        void put(int, double, double, double)
        int total_particles()
    
    cdef cppclass container_poly:
        container_poly(double,double,double,double,double,double,
                int,int,int,cbool,cbool,cbool,int) except +
        cbool compute_cell(voronoicell_neighbor &c, c_loop_all &vl)
        void put(int, double, double, double, double)
        int total_particles()
        
        
    cdef cppclass voronoicell_neighbor:
        voronoicell()
        void centroid(double &cx, double &cy, double &cz)
        double volume()
        double max_radius_squared()
        double total_edge_distance()
        double surface_area()
        double number_of_faces()
        double number_of_edges()
        
        void vertex_orders(vector[int]&)
        void vertices(double,double,double, vector[double]&)
        void face_areas(vector[double]&)
        void face_orders(vector[int] &)
        void face_freq_table(vector[int] &)
        void face_vertices(vector[int] &)
        void face_perimeters(vector[double] &)
        void normals(vector[double] &)
        void neighbors(vector[int] &)
    
    cdef cppclass c_loop_all:
        c_loop_all(container_base&)
        cbool start()
        cbool inc()
        int pid()
        void pos(double &x, double &y, double &z)
        

cdef class Cell:
    cdef voronoicell_neighbor *thisptr
    cdef int _id
    cdef double x,y,z
    cdef double r
    
    def __cinit__(self):
        self.thisptr = new voronoicell_neighbor()
    
    def __dealloc__(self):
        del self.thisptr
    
    @property
    def pos(self):
        return (self.x, self.y, self.z)
    
    @property
    def id(self):
        return self._id
    
    def volume(self): return self.thisptr.volume()
    def max_radius_squared(self): return self.thisptr.max_radius_squared()
    def total_edge_distance(self): return self.thisptr.total_edge_distance()
    def surface_area(self): return self.thisptr.surface_area()
    def number_of_faces(self): return self.thisptr.number_of_faces()
    def number_of_edges(self): return self.thisptr.number_of_edges()
    
    def centroid(self):
            cdef double cx = 0
            cdef double cy = 0
            cdef double cz = 0
            self.thisptr.centroid(cx,cy,cz)
            x,y,z = self.pos
            return (cx+x,cy+y,cz+z)
    
    def vertex_orders(self):
        cdef vector[int] v
        self.thisptr.vertex_orders(v)
        return v
    
    def vertices(self):
        cdef vector[double] v
        self.thisptr.vertices(self.x, self.y, self.z, v)
        return v
    
    def face_areas(self):
        cdef vector[double] v
        self.thisptr.face_areas(v)
        return v
    
    def face_freq_table(self):
        cdef vector[int] v
        self.thisptr.face_freq_table(v)
        return v
    
    def face_vertices(self):
        cdef vector[int] v
        self.thisptr.face_vertices(v)
        
        mylist = []
        
        it = iter(v)
        while True:
            try:
                n = next(it)
            except StopIteration:
                break
            mylist.append([next(it) for _ in range(n)])
        
        return mylist
    
    def face_perimeters(self):
        cdef vector[double] v
        self.thisptr.face_perimeters(v)
        return v
    
    def normals(self):
        cdef vector[double] v
        self.thisptr.normals(v)
        return list(zip(v[::3], v[1::3], v[2::3]))
    
    def neighbors(self):
        cdef vector[int] v
        self.thisptr.neighbors(v)
        return v
    
    def __str__(self):
        return '<Cell {0}>'.format(self._id)
    
    def __repr__(self):
        return '<Cell {0}>'.format(self._id)

cdef class Container:
    cdef container *thisptr
    def __cinit__(self, double ax_,double bx_,double ay_,double by_,double az_,double bz_,
                int nx_,int ny_,int nz_,cbool xperiodic_,cbool yperiodic_,cbool zperiodic_,int init_mem):
        self.thisptr = new container(ax_, bx_, ay_, by_, az_, bz_, nx_, ny_, nz_, 
                xperiodic_, yperiodic_, zperiodic_, init_mem)
    
    def __dealloc__(self):
        del self.thisptr
    
    def put(self, int n, double x, double y, double z):
        self.thisptr.put(n,x,y,z)
    
    def get_cells(self):
        cdef container_base *baseptr = (<container_base *>(self.thisptr))
        cdef c_loop_all *vl = new c_loop_all(dereference(baseptr))
        
        cell = Cell()
        
        cdef int vcells_left = self.thisptr.total_particles()
        cdef int id
        
        mylist = [None for _ in range(vcells_left)]
        
        if not vl.start():
            del vl
            raise ValueError("Failed to start loop")
        
        while True:
            if(self.thisptr.compute_cell(dereference(cell.thisptr), dereference(vl))):
                cell._id = vl.pid()
                assert(cell._id <= self.thisptr.total_particles())
                
                vl.pos(cell.x,cell.y,cell.z)
                mylist[cell._id] = cell
                
                vcells_left -= 1
                cell = Cell()
            if not vl.inc(): break
        
        del vl
        
        if vcells_left != 0:
            raise ValueError("Computation failed")
        return mylist

cdef class ContainerPoly:
    cdef container_poly *thisptr
    def __cinit__(self, double ax_,double bx_,double ay_,double by_,double az_,double bz_,
                int nx_,int ny_,int nz_,cbool xperiodic_,cbool yperiodic_,cbool zperiodic_,int init_mem):
        self.thisptr = new container_poly(ax_, bx_, ay_, by_, az_, bz_, nx_, ny_, nz_, 
                xperiodic_, yperiodic_, zperiodic_, init_mem)
    
    def __dealloc__(self):
        del self.thisptr
    
    def put(self, int n, double x, double y, double z, double r):
        self.thisptr.put(n,x,y,z,r)
    
    def get_cells(self):
        cdef container_base *baseptr = (<container_base *>(self.thisptr))
        cdef c_loop_all *vl = new c_loop_all(dereference(baseptr))
        
        cell = Cell()
        
        cdef int vcells_left = self.thisptr.total_particles()
        cdef int id
        
        mylist = [None for _ in range(vcells_left)]
        
        if not vl.start():
            del vl
            raise ValueError("Failed to start loop")
        
        while True:
            if(self.thisptr.compute_cell(dereference(cell.thisptr), dereference(vl))):
                cell._id = vl.pid()
                assert(cell._id <= self.thisptr.total_particles())
                
                vl.pos(cell.x,cell.y,cell.z)
                mylist[cell._id] = cell
                
                vcells_left -= 1
                cell = Cell()
            if not vl.inc(): break
        
        del vl
        
        if vcells_left != 0:
            raise ValueError("Computation failed")
        return mylist



from ._voro import Container as _Container, ContainerPoly as _ContainerPoly, Cell

class Container(list):
    def __init__(self, points, limits=1.0, periodic=False, radii=None, blocks=None):
        """Get the voronoi cells for a given set of points.
        
        points: an Nx3 iterable of floating point numbers denoting the coordinates.
        limits: the size of the box. May be a number (for a cubic box), or a 3-tuple.
        periodic: a bool or 3-tuple of bools representing wall periodicity
        """
        # make px, py, pz from periodic, whether periodic is a 3-tuple or bool
        try:
            px, py, pz = periodic
        except TypeError:
            px = py = pz = periodic
        px = bool(periodic)
        py = bool(periodic)
        pz = bool(periodic)
        
        # make lx, ly, lz from limits, whether limits is a 3-tuple or float
        try:
            lx, ly, lz = limits
        except TypeError:
            lx = ly = lz = limits
        lx = float(lx)
        ly = float(ly)
        lz = float(lz)
        assert lx > 0
        assert ly > 0
        assert lz > 0
        
        N = len(points)
        
        # make bx, by, bz from blocks, or make it up
        if blocks is None:
            Nthird = pow(N, 1.0/3.0)
            blocks = round(Nthird / lx), round(Nthird / ly), round(Nthird / lz)
        
        try:
            bx, by, bz = blocks
        except TypeError:
            bx = by = bz
        bx = max(int(bx), 1)
        by = max(int(by), 1)
        bz = max(int(bz), 1)
        
        # If we have periodic conditions, we want to get the 'boxed' version of each position.
        # Each coordinate (x,y,z) may or may not be periodic, so we'll deal with them separately.
        def roundedoff(n, l, periodic):
            if periodic:
                return float(n) % l
            else:
                return float(n)
        
        # voro has two types: Container and ContainerPoly. ContainerPoly is for unequal radii,
        # Container is for no-radii.
        # Now we choose the right one.
        if radii is not None:
            assert(len(radii) == len(points))
            self._container = _ContainerPoly(0,lx, 0,ly, 0,lz, # limits
                                bx, by, bz,        # block size
                                px, py, pz,        # periodicity
                                len(points))
            
            for n,(x,y,z),r in zip(range(len(points)), points, radii):
                self._container.put(n, roundedoff(x,lx,px), roundedoff(y,ly,py), roundedoff(z,lz,pz), float(r))
        else:
            # no radii => use voro._Container
            self._container = _Container(0,lx, 0,ly, 0,lz,         # limits
                                    bx, by, bz,        # block size
                                    px, py, pz,        # periodicity
                                    len(points))
            for n,(x,y,z) in enumerate(points):
                self._container.put(n, roundedoff(x,lx,px), roundedoff(y,ly,py), roundedoff(z,lz,pz))
            
        cells = self._container.get_cells()
        list.__init__(self, cells)
        
        # Sometimes a _Container has calculation issues. That can lead to the following.
        if len(self) != len(points):
            raise ValueError("Points could not be suitably fitted into the given box. \n\
You may want to check that all points are within the box, and none are overlapping. {} / {}".format(len(self), len(points)))
    
    def get_widths(self):
        return self._container.get_widths()
        
    def _get_bond_normals(self):
        """Returns a generator of [(dx,dy,dz,A) for each bond] for each cell.
        
        (dx,dy,dz) is the normal, and A is the area of the voronoi face."""
        return [
            [(x,y,z,A) for (x,y,z), A in zip(vc.normals(), vc.face_areas())]
            for vc in self
        ]

    def order(self, l=6, local=False, weighted=True):
        """Returns crystalline order parameter.
        
        Requires numpy and scipy.
        
        l is the which Q you want (6 is standard, for detecting hexagonal lattices)
        local is whether to calculate the Local Q6 or Global Q6
        weighted is whether or not to weight by area of the faces of each polygonal side
        """
        import numpy as np
        
        bonds = self._get_bond_normals()
        if not local:
            bonds = [np.concatenate(bonds)]
        
        Qs = []
        for blst in bonds:
            blst = np.array(blst)
            xyz, A = blst[:,:3], blst[:,3]
            Qs.append(orderQ(l, xyz, weights=(A if weighted else 1)))
        return np.mean(Qs)

def cart_to_spher(xyz):
    """Takes Nx3 matrix of Cartesian coordinates, and converts them to (theta, phi).
    
    Requires numpy.
    
    Returns an Nx2 matrix.
    
    Column 0: the "elevation" angle, 0 to pi
    Column 1: the "azimuthal angle, 0 to 2*pi
    """
    import numpy as np
    ptsnew = np.zeros((xyz.shape[0],2))
    xy = xyz[:,0]**2 + xyz[:,1]**2
    ptsnew[:,0] = np.arctan2(np.sqrt(xy), xyz[:,2]) # for elevation angle defined from Z-axis down
    ptsnew[:,1] = np.arctan2(xyz[:,1], xyz[:,0])
    return ptsnew

def orderQ(l, xyz, weights=1):
    """Returns Ql, for a given l (int) and a set of Cartesian coordinates xyz.
    
    Requires numpy and scipy.
    
    For global Q6, use l=6, and pass xyz of all the bonds.
    For local Q6, use l=6, and pass xyz of the bonds of each atom separately, and then average the results.
    
    To weight by Voronoi neighbors, pass weights=(face areas).
    """
    import numpy as np
    from scipy.special import sph_harm
    
    theta, phi = cart_to_spher(xyz).T
    weights = np.ones(xyz.shape[0]) * weights
    weights /= np.sum(weights)
    mmeans = np.zeros(2*l+1, dtype=float)
    for m in range(-l, l+1):
        sph_weighted = sph_harm(m, l, phi, theta).dot(weights)
        mmeans[m] = abs(sph_weighted)**2
    return np.sqrt(4*np.pi/(2*l+1) * np.sum(mmeans))



from distutils.core import setup
from distutils.extension import Extension
from distutils.command.sdist import sdist as _sdist
import os.path

try:
    from Cython.Build import cythonize
except ImportError:
    cythonize = None

extension_version = '0.1.2'

# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

if cythonize is not None:
    print("Building with Cython.")
    ext = cythonize(Extension("tess._voro",
              sources=["tess/_voro.pyx", "src/voro++.cc"],
              include_dirs=["src"],
              language="c++",
              ))
else:
    print("Cython not found, using prebuilt file.")
    ext = [Extension("tess._voro",
              sources=["tess/_voro.cpp", "src/voro++.cc"],
              include_dirs=["src"],
              language="c++",
              )]

# Set sdist to make the .cpp file
# from http://stackoverflow.com/a/18418524/4190270
class sdist(_sdist):
    def run(self):
        # this is already imported, but the import might have failed. If so, raise an ImportError now.
        from Cython.Build import cythonize
        
        # Make sure the compiled Cython files in the distribution are up-to-date
        cythonize(['tess/_voro.pyx'])
        _sdist.run(self)

cmdclass = dict(sdist = sdist)

# create the extension and add it to the python distribution
setup( name='tess', 
        version=extension_version, 
        author="Wendell Smith",
        author_email="wackywendell@gmail.com",
        description = ("A module for calculating and analyzing Voronoi tesselations"),
        license = "BSD",
        keywords = "laguerre voronoi tesselation voro++",
        url = "https://github.com/wackywendell/tess",
        long_description=read('README.rst'),
        classifiers=[
            "Development Status :: 3 - Alpha",
            "Topic :: Scientific/Engineering :: Mathematics",
            "Topic :: Scientific/Engineering :: Physics",
            "Intended Audience :: Science/Research",
            "Operating System :: POSIX :: Linux",
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: BSD License",
        ],
        
        packages=['tess'],
        package_dir={"tess": "tess"},
        
        ext_modules = ext,
        cmdclass=cmdclass,
        )
